#!/usr/bin/env python2

# -*- coding: utf-8 -*-
"""
@author: Szabone Andre Karolina

This program inserts bad pixel mask from a file (e.g., 'bad_pixel_mask.bmp')
    into a UFO Capture mask file.
The new mask file will be saved with an extension '+badpixmask.bmp' in the
    directory of the original mask file. If the original mask file was
    'MyMask.bmp' the the new file will be 'MyMask+badpixmask.bmp'.
All actions will be logged in the log file
    'logfile_UFO_add_bad_pixel_mask_to_base_mask.txt'
    in the directory of the original mask file.

Usage: If python is integrated in your system, you can run this module from a
       consol/command window or from a batch file simply by typing
       UFO_add_bad_pixel_mask_to_base_mask.py [C:\full\path\to\UFOCapture_mask_file.bmp  C:\full\path\to\bad_pixel_mask.bmp]

       If optional command line arguments are omitted the code runs
           interactively. Just run the module and answer the questions.
       The program will ask first for the full path and name of the mask file
           to be converted.
       Then the program asks for the full path and name of the file which
           contains the bad pixel mask (usually named 'bad_pixel_mask.bmp')

Note: This code has been written and tested under python 2.7.6
      This code was built using the datetime and pillow (2.4.0) packages

Runtime environment:

This program was written using WinPython-32bit-2.7.6.4. This distribution
runs under Windows XP and above. The latest Windows distribution it was tested
on was Windows 7. WinPython-32bit-2.7.6.4 contains all necessary components to
run this code. If you want to integrate python from this distribution in your
system, run 'WinPython Control Panel.exe' after installation, and select the
menu item 'Advanced'=> 'Register distribution...'

Concept:

A mask file can be set in UFO Capture to exclude areas from event detection.
This code serves to add the mask of bad pixels to such mask files so that
those pixels will be masked also in ...M.bmp files generated for each record.
Bad pixel mask files are a black and white BMP images with unwanted (bad)
    pixels in white. The size of the image must match the size of the mask
    file to be converted.

Information in UFO Capture mask files:

Masked pixels have RGB values (255,0,0). Otherwise the pixel is not masked.

Processing of the UFO Capture mask file:

To mask a pixel, its RGB value is set to (255,0,0).


"""
#### Importing general modules
from PIL import Image
import os
import sys
import datetime
import time

# Setting logfile
module_path = os.path.dirname(os.path.realpath(__file__))

def Logging(event, logfile, printing):
    # Write the log file
    with open(logfile, "a") as log_file:
        log_file.write(event+"\n")
    if printing:
        print event
    return


def main_prog(base_mask_name, bad_pixel_mask_name, logfile):
    
    # Checking if the files are valid (exist and have right extension)
    error_1 = CheckValidityOfFile(base_mask_name)
    error_2 = CheckValidityOfFile(bad_pixel_mask_name)
    if error_1 or error_2:
        return

    # Opening image files
    base_image = Image.open(base_mask_name)
    bp_image = Image.open(bad_pixel_mask_name)
    
    # Checking image sizes
    base_width, base_height = base_image.size
    bp_width, bp_height = bp_image.size
    if ((base_width != bp_width) or (base_height != bp_height)):
        Logging("Sizes of the two files does not match",
                    logfile, printing=True)
        return
    else:
        new_filename = base_mask_name[:-4]+"+badpixmask.bmp"
        PutBadPixelsToBaseMask(base_image, bp_image, new_filename, logfile)
    return


def CheckValidityOfFile(file_name):
    """
    """
    error = False
    if (os.path.isfile(base_mask_name)):
            if (base_mask_name[-4:] != ".bmp"):
                Logging("Invalid file extension of base mask (must be .bmp)",
                        logfile, printing=True)
                error = True
    else:
        Logging("File does not exist: "+base_mask_name, logfile, printing=True)
        error = True
    return error
    
    
def PutBadPixelsToBaseMask(base_image, bp_image, new_filename, logfile):
    """
    """
    r_base, g_base, b_base = base_image.split()
    new_r = r_base.point(lambda i: 255)
    new_g = g_base.point(lambda i: 0)
    new_b = b_base.point(lambda i: 0)
    r_base.paste(new_r, None, bp_image)
    g_base.paste(new_g, None, bp_image)
    b_base.paste(new_b, None, bp_image)
    
    new_im = Image.merge(base_image.mode, (r_base,g_base,b_base))
    new_im.save(new_filename, "BMP")
    Logging("Successfully saved new base mask: "+new_filename,
            logfile, printing=True)
    return


if (len(sys.argv) == 3):
    base_mask_name = sys.argv[1]
    logfile_path = '\\'.join(base_mask_name.split("\\")[:-1]) + '\\'
    logfile = logfile_path+'logfile_UFO_add_bad_pixel_mask_to_base_mask.txt'
    Logging("---------------------\n START PROGRAM \n"+str(datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S'))+"\n---------------------", 
        logfile, printing=True)    
    Logging("The absolute path and name of the base mask file to be modified: "+base_mask_name,
        logfile, printing=True)
    bad_pixel_mask_name = sys.argv[2]
    Logging("The absolute path and name of the bad pixel mask: "+bad_pixel_mask_name,
            logfile, printing=True)
else:
    base_mask_name = raw_input("Give the absolute path and name of the base mask file to be modified:\n")
    logfile_path = '\\'.join(base_mask_name.split("\\")[:-1]) + '\\'
    logfile = logfile_path+'logfile_UFO_add_bad_pixel_mask_to_base_mask.txt'
    Logging("---------------------\n START PROGRAM \n"+str(datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S'))+"\n---------------------", 
            logfile, printing=False)
    Logging("Give the absolute path and name of the base mask file to be modified:\n"+base_mask_name,
            logfile, printing=False)
    bad_pixel_mask_name = raw_input("Give the absolute path and name of the bad pixel mask:\n")
    Logging("Give the absolute path and name of the bad pixel mask:\n"+bad_pixel_mask_name,
            logfile, printing=False)


main_prog(base_mask_name, bad_pixel_mask_name, logfile)

Logging("---------------------\n END OF PROGRAM \n"+str(datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S'))+"\n---------------------", 
        logfile, printing=True)