#!/usr/bin/env python2

# -*- coding: utf-8 -*-
"""
@author: Szabone Andre Karolina

This code processes ...M.bmp mask (M) files generated by UFO Capture.
Scintillating pixels present in all M files in a given directory are
    identified. Those pixels are most probably bad pixels.
A mask file named 'bad_pixel_mask.bmp' is created from these pixels and will be
    saved in the given directory.
If requested, an ...N.bmp (N) file is created from each M file in
    the given directory using 'bad_pixel_mask.bmp' file in the given directory.
    N files are used by UFO Analyzer. With bad pixels masked out, fitting of
    the star map to detected scintillating objects is easier.
All actions will be logged in the given directory in a file named
    'logfile_UFO_bad_pixel_mask_processor.txt'.

Usage: the code is interactive. Just run this module in your favourite python
       interpreter and answer the questions.
       If python is integrated in your system, you can run this module from a
       consol/command window simply by typing 'UFO_bad_pixel_mask_processor.py'

       The program will ask for the full path of a directory it will use.
       If a 'bad_pixel_mask.bmp' file is to be created, the directory must
       contain at least 2 M files.

Note: This code has been written and tested under python 2.7.6
      This code was built using the datetime and pillow (2.4.0) packages

Runtime environment:

This program was written using WinPython-32bit-2.7.6.4. This distribution
runs under Windows XP and above. The latest Windows distribution it was tested
on was Windows 7. WinPython-32bit-2.7.6.4 contains all necessary components to
run this code. If you want to integrate python from this distribution in your
system, run 'WinPython Control Panel.exe' after installation, and select the
menu item 'Advanced'=> 'Register distribution...'

Concept:

Stars are registered as scintillating objects in ...M.bmp mask files saved by
UFO Capture for each record. The position of stars changes over time. If a
pixel is registered as scintillating in all records taken at different time, it
is most probably a pixel does not mark a star. The presence of such pixels is
annoying when the direction of the camera is to be determined by star map
matching in the UFO analyzer programme.
bad_pixel_mask.bmp: a black and white image with unwanted (bad) pixels in
    white. If a pixel was identified as 'bad', its surrounding pixels are also
    marked as 'bad'. The size of the image matches the size of the M files
    it was deduced from.

Information in M files:

In M files, the information is coded in the color channels.
Note that the 2 pixel wide frame of the image is treated differently.
If red value is nonzero, the pixel is one of the record-triggering pixels
If blue value is 0, it is a general background pixel (opt-mask).
If blue value is 2 (or 1 in the frame), the pixel is masked (area-mask).
If blue value is 250 (or 249 in the frame), the pixel is scintillating
    but it is also masked (area-mask + scl-mask).
If blue value is 248, the pixel is scintillating, not masked (scl-mask).

Creation of N file from an M file:

- around red pixels (+-10 pixel range in both x and y directions):
    - if blue value < 128 => not changed
    - if blue value > 128 => 8 is subtracted from blue value to add area-mask
- outside the square region of red pixels:
    - if blue value < 128 => 4 is added to blue value
    - if blue value > 128 => 4 is subtracted from blue value to add area-mask
- +1 is added to the blue value in the frame

"""
#### Importing general modules
from PIL import Image, ImageChops
import os
import datetime
import time

# Setting logfile
module_path = os.path.dirname(os.path.realpath(__file__))


def Logging(event, logfile, printing):
    # Write the log file
    with open(logfile, "a") as log_file:
        log_file.write(event+"\n")
    if printing:
        print event
    return 

def main_prog():    
    # Asking absolute path of the working directory from the user
    work_dir =  raw_input("Give the absolute path of the working directory (where M files are present):\n")
    logfile_path = work_dir + '\\'
    logfile = logfile_path+'logfile_UFO_bad_pixel_mask_processor.txt'
    
    Logging("---------------------\n START PROGRAM \n"+str(datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S'))+"\n---------------------", 
            logfile, printing=True)
    Logging("Give the absolute path of the working directory (where M files are present):",logfile, printing=False)
    Logging(work_dir,logfile,printing=False)
    if (not os.path.isdir(work_dir)):
        Logging("ERROR: directory does not exist: "+work_dir, logfile, printing=True)
        return
    
    # Creating list of the M.bmp files to be used in this program
    m_filenames = os.listdir(work_dir)
    m_filenames = [work_dir+"//"+s for s in m_filenames 
                                    if os.path.isfile(os.path.join(work_dir,s))]
    m_filenames = [s for s in m_filenames if s[-5:]=="M.bmp"]
    if (len(m_filenames) == 0):
        Logging("ERROR: there are no M.bmp file(s) in directory "+work_dir, logfile, printing=True)
        return
    
    # Deciding whether new bad_pixel_mask.bmp is needed or not
    generate_new_bad_pixel_mask = True
    bad_pixel_mask_name = work_dir+"//bad_pixel_mask.bmp"
    
    if (os.path.isfile(bad_pixel_mask_name)):
        while True:
            answer = raw_input("There is bad_pixel_mask.bmp in the working directory.\n\
Do you want to use it (U) or generate a new one (G)?\n\
(Warning: if you press G, the old bad_pixel_mask.bmp will be overwritten.)\n")
            Logging("There is bad_pixel_mask.bmp in the working directory.\n\
Do you want to use it (U) or generate a new one (G)?\n\
(Warning: if you press G, the old bad_pixel_mask.bmp will be overwritten.)", logfile, printing=False)
            Logging(answer,logfile,printing=False)
            if (answer == "U" or answer == "u"):
                generate_new_bad_pixel_mask = False
                break
            elif (answer == "G" or answer == "g"):
                break
            else:
                Logging("Invalid answer, try again", logfile, printing=True)
    
    # If needed, generating bad_pixel_mask.bmp
    if generate_new_bad_pixel_mask:
        # For generating new bad_pixel_mask.bmp at least two M.bmp files are needed
        if (len(m_filenames) < 2):
            Logging("ERROR: the number of M.bmp files is too small ("+\
                str(len(m_filenames))+") for generating bad_pixel_mask.bmp",\
                logfile, printing=True)
            return
        else:
            Logging("Generating bad_pixel_mask.bmp", logfile, printing=True)
            bad_pixel_mask = CreateBadPixelsMask(m_filenames, 
                                                 bad_pixel_mask_name,
                                                 logfile)
    else:
        Logging("Opening bad_pixel_mask.bmp", logfile, printing=True)
        bad_pixel_mask = Image.open(bad_pixel_mask_name)
    
    # Asking the user that N files are needed or not
    create_n_files = raw_input("Create N files from M files using bad_pixel_mask.bmp (Y/N)?\n\
(Warning: if Y, existing N files will be overwritten.)\n")
    Logging("Create N files from M files using bad_pixel_mask.bmp (Y/N)?\n\
(Warning: if Y, existing N files will be overwritten.)", logfile, printing=False)
    Logging(create_n_files, logfile, printing=False)
    if (create_n_files == "Y" or create_n_files == "y"):
        ConvertMfileToNfile(m_filenames, bad_pixel_mask, logfile)
    
    Logging("---------------------\n END OF PROGRAM \n"+str(datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S'))+"\n---------------------", 
        logfile, printing=True)
    return


def CreateBadPixelsMask(img_filenames, bad_pixel_mask_name, logfile):
    """
    Searches the "scintillating" pixels which are present on all of the images.
    Creates a mask to exclude them and their neighboring pixels from analysis
    with the UFOAnalyzer program.
    """
    # Size of area around a bad pixel (in pixels) 
    #  to be handled as area-mask (in UFO Analyzer)
    size_of_area = 1
    
    # Creating a base mask using the firs to-be-processed image file
    im = Image.open(img_filenames[0])
    r, g, b = im.split()
    
    # If the blue channel has the value of 248 (unmasked scintillating pixel),
    #  then change the value of the pixel to 255 otherwise setting it to 0
    #  The conversion to mode "1" is needed to use the ImageChops.logical_and function
    bad_pixel_mask_first = b.point(lambda i: i==248 and 255).convert("1")
    first_width, first_height = bad_pixel_mask_first.size
    
    # Looping trough the files and searching the pixels 
    #  which have value of 248 in the blue channel and
    #  which are in all of the files at the same location
    for ifile in img_filenames:
        Logging(ifile, logfile, printing=True)
        # Opening image file and splitting to channels
        im = Image.open(ifile)
        r, g, b = im.split()
        iwidth, iheight = im.size
        if ((iwidth != first_width) or (iheight != first_height)):
            Logging("WARNING: skipping file (sizes of the "+ifile+" does not match with the first file)",
                        logfile, printing=True)
        else:
            bad_pixel_mask_ifile = b.point(lambda i: i==248 and 255).convert("1")
            bad_pixel_mask_first = ImageChops.logical_and(bad_pixel_mask_first,
                                                      bad_pixel_mask_ifile)
    
    # Creating mask that has 0 values in every pixel
    bad_pixel_mask = b.point(lambda i: 0)

    # Looping trough the first bad pixel mask and extending it with size_of_area
    for ix in range(0,first_width):
        for iy in range(0,first_height):
            
            # If the value of the actual pixel is 255 (it is a bad pixel)
            if (bad_pixel_mask_first.getpixel((ix,iy))==255):
                # Get box coordinates paying attention to the edge of the image
                xmin = max((ix - size_of_area),0)
                xmax = min((ix + size_of_area + 1), first_width)
                ymin = max((iy - size_of_area),0)
                ymax = min((iy + size_of_area + 1), first_height)
                box = (xmin, ymin, xmax, ymax)
                # Cropping the box
                submask = bad_pixel_mask_first.crop(box)
                # Setting values of the bad pixel's neigborhood to 255
                submask = submask.point(lambda i: 255)
                # Pasing back to the mask
                bad_pixel_mask.paste(submask,box)
                
    # Saving the mask to a file
    bad_pixel_mask.save(bad_pixel_mask_name)
    
    Logging("Mask of bad pixels has been saved as "+bad_pixel_mask_name,\
                logfile, printing=True)
    return bad_pixel_mask


def ConvertMfileToNfile(img_filenames, bad_pixel_mask, logfile):
    """
    Converts M files to N files
    
    - near the red pixels (+-10 pixels):
     - if the value of the blue channel (BCV) is low: does not change the BCV
     - if it is high: subtracts 8 from the BCV
    - far from the red pixels:
     - if the BCV is low: adds 4 to the BCV
     - if it is high: subtracts 4 from the BCV
    - finally it adds 1 to the BCV at the edges of the image (2 pixel wide frame)
    """
    # Size of area around a red pixel (in pixels) to be handled as d-area (like in UFO Analyzer)
    size_of_area = 10
    
    # Looping trough the files
    for ifile in img_filenames:
        Logging(ifile, logfile, printing=True)
        new_filename = ifile[:-5]+"N.BMP"
        
        # Opening image file, splitting to channels and get image size
        im = Image.open(ifile)
        r, g, b = im.split()
        width, height = im.size
        bp_width, bp_height = bad_pixel_mask.size
        if (width != bp_width or height != bp_height):
            Logging("ERROR: size of the mask does not match to the M file's size",
                    logfile, printing=True)
        else:
            # Locating the red pixels (if the value of the pixel is 0 then return 0 else 255)
            red_mask = r.point(lambda i: i>0 and 255)
            
            # Looping trough the pixels and creating mask of the red channel
            for ix in range(0,width):
                for iy in range(0,height):
                    if r.getpixel((ix,iy))>0:
                        # Get box coordinates paying attention to the edge of the image
                        xmin = max((ix - size_of_area),0)
                        xmax = min((ix + size_of_area + 1), width)
                        ymin = max((iy - size_of_area),0)
                        ymax = min((iy + size_of_area + 1), height)
                        box = (xmin, ymin, xmax, ymax)
                        submask = red_mask.crop(box)
                        # Setting values of the red pixel's neigborhood to 255
                        submask = submask.point(lambda i: 255)
                        red_mask.paste(submask,box)
                        
            # Creating blue mask (high values: 255, low values: 0)
            blue_mask = b.point(lambda i: i > 128 and 255)
            
            # Creating other masks
            blue_mask = blue_mask.convert("1")
            red_mask = red_mask.convert("1")
            high_red_mask = ImageChops.logical_and(blue_mask, red_mask)
            high_nored_mask = ImageChops.logical_and(blue_mask, ImageChops.invert(red_mask))
            low_nored_mask = ImageChops.logical_and(ImageChops.invert(blue_mask), ImageChops.invert(red_mask))
            
            # Changing blue channel
            # low blue and no red: add 4
            new_b = b.point(lambda i: i+4)
            b.paste(new_b, None, low_nored_mask)
            # high blue and no red: subtract 4
            new_b = b.point(lambda i: i-4)
            b.paste(new_b, None, high_nored_mask)
            # high blue and red area: subtract 8    
            new_b = b.point(lambda i: i-8)
            b.paste(new_b, None, high_red_mask)
            
            # Correcting the pixels which are near the image's edge (2 px) and the
            # value of the blue channel is high: add 1
            edge_mask = b.point(lambda i: 255)
            box = (2,2,width-2, height-2)
            submask = edge_mask.crop(box)
            submask = submask.point(lambda i: 0)
            edge_mask.paste(submask,box)
            edge_mask = edge_mask.convert("1")
            edge_high_mask = ImageChops.logical_and(blue_mask,edge_mask)
            new_b = b.point(lambda i: i+1)
            b.paste(new_b, None, edge_high_mask)
            
            # Masking bad pixels
            new_b = b.point(lambda i: i+2)
            b.paste(new_b, None, bad_pixel_mask)
            
            # Merging and saving N file
            new_im = Image.merge(im.mode, (r,g,b))
            new_im.save(new_filename, "BMP")
            Logging("Successfully generated file: "+new_filename,
                    logfile, printing=True)
    return

#Logging("---------------------\n START PROGRAM \n---------------------", 
#        logfile, printing=True)
main_prog()
#Logging("---------------------\n END OF PROGRAM \n---------------------", 
#        logfile, printing=True)